"""Collection of mathematical operators."""
from .addition import Add, add
from .multiply import Mul, mul
from .negative import Neg, neg
from .power import Pow

from .joint import J
from .iid import Iid

from .trunkation import Trunc, Trunk, trunk

from .logarithm import Log
from .logarithm10 import Log10
from .logarithmn import Logn



"""Simple constructor to create single variables to create polynomials."""
import numpoly


def variable(dims=1):
    """
    Simple constructor to create single variables to create polynomials.

    Args:
        dims (int):
            Number of dimensions in the array.

    Returns:
        (chaospy.poly.polynomial):
            Polynomial array with unit components in each dimension.

    Examples:
        >>> chaospy.variable()
        polynomial(q0)
        >>> chaospy.variable(3)
        polynomial([q0, q1, q2])
    """
    return numpoly.symbols("q:%d" % dims)



"""Create an N-dimensional unit polynomial basis."""
import numpy
import numpoly


def basis(start, stop=None, dim=1, sort="G", cross_truncation=1.):
    """
    Create an N-dimensional unit polynomial basis.

    Args:
        start (int, numpy.ndarray):
            the minimum polynomial to include. If int is provided, set as
            lowest total order.  If array of int, set as lower order along each
            axis.
        stop (int, numpy.ndarray):
            the maximum shape included. If omitted:
            ``stop <- start; start <- 0`` If int is provided, set as largest
            total order. If array of int, set as largest order along each axis.
        dim (int):
            dim of the basis. Ignored if array is provided in either start or
            stop.
        sort (str):
            The polynomial ordering where the letters ``G``, ``I`` and ``R``
            can be used to set grade, inverse and reverse to the ordering.  For
            ``basis(start=0, stop=2, dim=2, order=order)`` we get:
            ======  ==================
            order   output
            ======  ==================
            ""      [1 y y^2 x xy x^2]
            "G"     [1 y x y^2 xy x^2]
            "I"     [x^2 xy x y^2 y 1]
            "R"     [1 x x^2 y xy y^2]
            "GIR"   [y^2 xy x^2 y x 1]
            ======  ==================
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        (chaospy.poly.ndpoly) : Polynomial array.

    Examples:
        >>> chaospy.basis(4, 4, 2, sort="GR")
        polynomial([q0**4, q0**3*q1, q0**2*q1**2, q0*q1**3, q1**4])
        >>> chaospy.basis([1, 1], [2, 2], sort="GR")
        polynomial([q0*q1, q0**2*q1, q0*q1**2, q0**2*q1**2])
    """
    if stop is None:
        start, stop = 0, start
    dim = max(numpy.asarray(start).size, numpy.asarray(stop).size, dim)
    return numpoly.monomial(
        start=start,
        stop=numpy.array(stop)+1,
        ordering=sort,
        cross_truncation=cross_truncation,
        names=numpoly.symbols("q:%d" % dim, asarray=True),
    )



"""Adjust the dimensions of a polynomial."""
import numpoly


def setdim(poly, dim=None):
    """
    Adjust the dimensions of a polynomial.

    Output the results into ndpoly object

    Args:
        poly (chaospy.poly.ndpoly):
            Input polynomial
        dim (int):
            The dimensions of the output polynomial. If omitted, increase
            polynomial with one dimension. If the new dim is smaller then
            `poly`'s dimensions, variables with cut components are all cut.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> poly = x*x-x*y
        >>> chaospy.setdim(poly, 1)
        polynomial(q0**2)
        >>> chaospy.setdim(poly, 3)
        polynomial(q0**2-q0*q1)
        >>> chaospy.setdim(poly, 3).names
        ('q0', 'q1', 'q2')
    """
    poly = numpoly.polynomial(poly)
    indices = [int(name[1:]) for name in poly.names]
    dim = max(indices)+2 if dim is None else dim
    poly = poly(**{("q%d" % index): 0 for index in indices if index >= dim})
    _, poly = numpoly.align_indeterminants(numpoly.symbols("q:%d" % dim), poly)
    return poly



"""
Constructing a multivariate probability distribution consisting of identical
independent distributed marginals can be done using the
:func:`~chaospy.distributions.operators.iid.Iid`. E.g.::

    >>> X = chaospy.Normal()
    >>> Y = chaospy.Iid(X, 4)
    >>> print(Y.sample())
    [ 0.39502989 -1.20032309  1.64760248 -0.04465437]
"""
from copy import deepcopy
from .joint import J


class Iid(J):
    """
    Opaque method for creating independent identical distributed random
    variables from an univariate variable.

    Args:
        dist (Dist):
            Distribution to make into i.i.d. vector.
    """

    def __init__(self, dist, length):
        assert len(dist) == 1 and length >= 1
        J.__init__(self, *[deepcopy(dist) for _ in range(length)])

    def __str__(self):
        return (self.__class__.__name__ + "(" + str(self.prm["_000"]) +
                ", " + str(len(self)) + ")")



"""Main Sobol sensitivity index."""
import numpy

from ...poly.setdim import setdim
from ..variance import Var
from ..conditional import E_cond


def Sens_m(poly, dist, **kws):
    """
    Variance-based decomposition/Sobol' indices.

    First order sensitivity indices.

    Args:
        poly (chaospy.poly.ndpoly):
            Polynomial to find first order Sobol indices on.
        dist (Dist):
            The distributions of the input used in ``poly``.

    Returns:
        (numpy.ndarray):
            First order sensitivity indices for each parameters in ``poly``,
            with shape ``(len(dist),) + poly.shape``.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, x, y, 10*x*y])
        >>> dist = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> indices = chaospy.Sens_m(poly, dist)
        >>> print(indices)
        [[0.         1.         0.         0.42857143]
         [0.         0.         1.         0.42857143]]
    """
    dim = len(dist)
    poly = setdim(poly, dim)

    zero = [0]*dim
    out = numpy.zeros((dim,) + poly.shape)
    V = Var(poly, dist, **kws)
    for i in range(dim):
        zero[i] = 1
        out[i] = Var(E_cond(poly, zero, dist, **kws),
                     dist, **kws)/(V+(V == 0))*(V != 0)
        zero[i] = 0
    return out



r"""
A simple polynomial can be created through ``variable`` constructor. For
example to construct a simple bivariate polynomial::

    >>> q0, q1 = chaospy.variable(dims=2)
    >>> q0
    polynomial(q0)

A collection of polynomial can be manipulated using basic arithmetic operators
and joined together into polynomial expansions::

    >>> polys = chaospy.polynomial([1, q0, q0*q1])
    >>> polys
    polynomial([1, q0, q0*q1])

Note that constants and simple polynomials can be joined together into arrays
without any problems.

In practice, having the ability to fine tune a polynomial exactly as one
wants it can be useful, but it can also be cumbersome when dealing with
larger arrays for application. To automate the construction of simple
polynomials, there is the ``basis`` constructor. In its simplest forms
it creates an array of simple monomials::

    >>> chaospy.basis(start=4)
    polynomial([1, q0, q0**2, q0**3, q0**4])

It can be expanded to include number of dimensions and a lower bound for the
polynomial order::

    >>> chaospy.basis(start=1, stop=2, dim=2)
    polynomial([q1, q0, q1**2, q0*q1, q0**2])

There is also the possibility to create anisotropic expansions::

    >>> chaospy.basis(start=1, stop=[1, 2])
    polynomial([q1, q0, q1**2, q0*q1, q0*q1**2])

and the possibility to fine tune the sorting of the polynomials::

    >>> chaospy.basis(start=1, stop=2, dim=2, sort="GRI")
    polynomial([q1**2, q0*q1, q0**2, q1, q0])

To manipulate a polynomial there is a collection of tools available. In
table [tab\_poly] they are listed with description. Much of the behavior
is overlapping with the numerical library ``numpy``. The functions that
overlaps in name is backwards compatible, e.g. if anything else than a
polynomial is inserted into ``chaospy.sum``, the argument is passed to
``numpy.sum``.

Any constructed polynomial is a callable. The argument can either be
inserted positional or as keyword arguments ``q0``, ``q1``, ...::

    >>> poly = chaospy.polynomial([1, q0**2, q0*q1])
    >>> poly(2, 3)
    array([1, 4, 6])
    >>> poly(q1=3, q0=2)
    array([1, 4, 6])

The input can be a mix of scalars and arrays, as long as the shapes together
can be joined to gether in a common compatible shape::

    >>> poly(2, [1, 2, 3, 4])
    array([[1, 1, 1, 1],
           [4, 4, 4, 4],
           [2, 4, 6, 8]])

It is also possible to perform partial evaluation, i.e. evaluating some of the
dimensions. This can be done using the keyword arguments::

    >>> poly(q0=2)
    polynomial([1, 4, 2*q1])
    >>> poly(q1=2)
    polynomial([1, q0**2, 2*q0])

The type of return value for the polynomial is ``numpy.ndarray`` if all
dimensions are filled. If not, it returns a new polynomial.

In addition to input scalars, arrays and masked values, it is also possible to
pass simple polynomials as argument. This allows for variable substitution. For
example, to swap two dimensions, one could simply do the following::

    >>> poly(q1, q0)
    polynomial([1, q1**2, q0*q1])

It is also possible to do all the over mentioned methods together in the same
time. For example, partial evaluation and variable substitution::

    >>> poly(q1=q1**3-1)
    polynomial([1, q0**2, -q0+q0*q1**3])
"""
from functools import wraps

from .basis import basis
from .prange import prange
from .variable import variable
from .setdim import setdim

from numpoly import (
    abs, absolute, add, any, all, allclose, around, aspolynomial, atleast_1d,
    atleast_2d, atleast_3d, call, ceil, concatenate, cumsum, decompose, diff,
    divide, dstack, equal, floor, gradient, hessian, hstack, inner, isclose,
    isconstant, isfinite, multiply, ndpoly, negative, not_equal, outer,
    polynomial, positive, power, prod, repeat, rint, round, square, stack,
    subtract, sum, vstack,
)

@wraps(polynomial)
def Poly(*args, **kwargs):
    print("Deprecation warning: "
          "`chaospy.Poly` replaced by `chaospy.polynomial`")
    return polynomial(*args, **kwargs)



"""
Clenshaw-Curtis quadrature method is a good all-around quadrature method
comparable to Gaussian quadrature, but typically limited to finite intervals
without a specific weight function. In addition to be quite accurate, the
weights and abscissas can be calculated quite fast.

Another thing to note is that Clenshaw-Curtis, with an appropriate growth rule
is fully nested. This means, if one applies a method that combines different
order of quadrature rules, the number of evaluations can often be reduced as
the abscissas can be used across levels.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="clenshaw_curtis")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0. 1.]] [0.5 0.5]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.   0.25 0.75 1.  ]] [0.056 0.444 0.444 0.056]

The first few orders with exponential growth rule::

    >>> for order in [0, 1, 2]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="clenshaw_curtis", growth=True)
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    2 [[0.    0.146 0.5   0.854 1.   ]] [0.033 0.267 0.4   0.267 0.033]

Applying the rule using Smolyak sparse grid::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     2, distribution, rule="clenshaw_curtis",
    ...     growth=True, sparse=True)
    >>> abscissas.round(2)
    array([[0.  , 0.  , 0.  , 0.15, 0.5 , 0.5 , 0.5 , 0.5 , 0.5 , 0.85, 1.  ,
            1.  , 1.  ],
           [0.  , 0.5 , 1.  , 0.5 , 0.  , 0.15, 0.5 , 0.85, 1.  , 0.5 , 0.  ,
            0.5 , 1.  ]])
    >>> weights.round(3)
    array([ 0.028, -0.022,  0.028,  0.267, -0.022,  0.267, -0.089,  0.267,
           -0.022,  0.267,  0.028, -0.022,  0.028])
"""
from __future__ import division

import numpy

from .combine import combine_quadrature


def quad_clenshaw_curtis(order, domain, growth=False):
    """
    Generate the quadrature nodes and weights in Clenshaw-Curtis quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Dist, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_clenshaw_curtis(3, (0, 1))
        >>> abscissas.round(4)
        array([[0.  , 0.25, 0.75, 1.  ]])
        >>> weights.round(4)
        array([0.0556, 0.4444, 0.4444, 0.0556])
    """
    from ..distributions.baseclass import Dist
    if isinstance(domain, Dist):
        abscissas, weights = quad_clenshaw_curtis(
            order, (domain.lower, domain.upper), growth)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = numpy.array(domain)
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    if growth:
        order = numpy.where(order > 0, 2**order, 0)

    abscissas, weights = zip(*[_clenshaw_curtis(order_) for order_ in order])

    return combine_quadrature(abscissas, weights, (lower, upper))


def _clenshaw_curtis(order):
    r"""
    Backend method.

    Examples:
        >>> abscissas, weights = _clenshaw_curtis(0)
        >>> abscissas
        array([0.5])
        >>> weights
        array([1.])
        >>> abscissas, weights = _clenshaw_curtis(1)
        >>> abscissas
        array([0., 1.])
        >>> weights
        array([0.5, 0.5])
        >>> abscissas, weights = _clenshaw_curtis(2)
        >>> abscissas
        array([0. , 0.5, 1. ])
        >>> weights
        array([0.16666667, 0.66666667, 0.16666667])
        >>> abscissas, weights = _clenshaw_curtis(3)
        >>> abscissas
        array([0.  , 0.25, 0.75, 1.  ])
        >>> weights
        array([0.05555556, 0.44444444, 0.44444444, 0.05555556])
        >>> abscissas, weights = _clenshaw_curtis(4)
        >>> abscissas
        array([0.        , 0.14644661, 0.5       , 0.85355339, 1.        ])
        >>> weights
        array([0.03333333, 0.26666667, 0.4       , 0.26666667, 0.03333333])
    """
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscisas = 0.5*numpy.cos(theta) + 0.5

    idx, idy = numpy.mgrid[:order+1, :order//2]
    weights = 2*numpy.cos(2*(idy+1)*theta[idx])/(4*idy*(idy+2)+3)
    if order % 2 == 0:
        weights[:, -1] *= 0.5
    weights = (1-numpy.sum(weights, -1)) / order

    weights[0] /= 2
    weights[-1] /= 2

    return abscisas, weights



"""
Uncertainty Quantification Toolbox
==================================

This module contains tools for performing uncertainty quantification of models.
"""
import logging
import os
import pkg_resources

import chaospy.bertran
import chaospy.chol
import chaospy.descriptives
import chaospy.distributions
import chaospy.orthogonal
import chaospy.poly
import chaospy.spectral
import chaospy.quadrature
import chaospy.saltelli
import chaospy.regression

from chaospy.distributions import *
from chaospy.orthogonal import *
from chaospy.poly import *
from chaospy.spectral import *
from chaospy.quadrature import *
from chaospy.saltelli import *
from chaospy.descriptives import *
from chaospy.regression import *
from chaospy.external import *

LOGPATH = os.environ.get("CHAOSPY_LOGPATH", os.devnull)
logging.basicConfig(level=logging.DEBUG, filename=LOGPATH, filemode="w")
streamer = logging.StreamHandler()
streamer.setLevel(logging.WARNING)
logger = logging.getLogger(__name__)
logger.addHandler(streamer)

try:
    __version__ = pkg_resources.get_distribution("chaospy").version
except pkg_resources.DistributionNotFound:
    pass
